from __future__ import annotations
from typing import TYPE_CHECKING
from ..mock import mock_g

if mock_g.DOCS_BUILDING:
    from ..mock import unohelper
else:
    import unohelper

from com.sun.star.awt import XTopWindowListener

if TYPE_CHECKING:
    from com.sun.star.lang import EventObject


class XTopWindowAdapter(unohelper.Base, XTopWindowListener):
    """
    makes it possible to receive window events.

    This class is meant a parent class.
    """

    def windowOpened(self, event: EventObject) -> None:
        """is invoked when a window is activated."""
        pass

    def windowActivated(self, event: EventObject) -> None:
        """is invoked when a window is activated."""
        pass

    def windowDeactivated(self, event: EventObject) -> None:
        """is invoked when a window is deactivated."""
        pass

    def windowMinimized(self, event: EventObject) -> None:
        """is invoked when a window is iconified."""
        pass

    def windowNormalized(self, event: EventObject) -> None:
        """is invoked when a window is deiconified."""
        pass

    def windowClosing(self, event: EventObject) -> None:
        """
        is invoked when a window is in the process of being closed.

        The close operation can be overridden at this point.
        """
        pass

    def windowClosed(self, event: EventObject) -> None:
        """is invoked when a window has been closed."""
        pass

    def disposing(self, event: EventObject) -> None:
        """
        gets called when the broadcaster is about to be disposed.

        All listeners and all other objects, which reference the broadcaster
        should release the reference to the source. No method should be invoked
        anymore on this object ( including XComponent.removeEventListener() ).

        This method is called for every listener registration of derived listener
        interfaced, not only for registrations at XComponent.
        """
        # from com.sun.star.lang.XEventListener
        pass



from enum import IntEnum
from . import kind_helper


class SearchMatchKind(IntEnum):
    """Search Match Kind"""

    FULL = 1
    """Full search match. Search criteria must match Exact"""
    FULL_IGNORE_CASE = 2
    """Full search match. Case is ignored. Search criteria must match Exact"""
    PARTIAL = 3
    """Partial match. Search criteria must be in the search result"""
    PARTIAL_IGNORE_CASE = 4
    """Partial match. Case is ignored. Search criteria must be in the search result"""

    @staticmethod
    def from_str(s: str) -> "SearchMatchKind":
        """
        Gets an ``SearchMatchKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``SearchMatchKind`` instance.

        Returns:
            SearchMatchKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, SearchMatchKind)



from enum import IntEnum
from . import kind_helper


class GluePointsKind(IntEnum):
    """Glue Point Kind"""

    TOP = 0
    RIGHT = 1
    BOTTOM = 2
    LEFT = 3

    @staticmethod
    def from_str(s: str) -> "GluePointsKind":
        """
        Gets an ``GluePointsKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``GluePointsKind`` instance.

        Returns:
            GluePointsKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, GluePointsKind)



# coding: utf-8
"""
Calc Named Events.
"""
from __future__ import annotations
from typing import NamedTuple


class CalcNamedEvent(NamedTuple):
    """
    Named events for :py:class:`~.office.calc.Calc` class
    """

    DOC_OPENING = "calc_doc_opening"
    """Doc Opening Spreadsheet see :py:meth:`Calc.open_doc() <.office.calc.Calc.open_doc>`"""
    DOC_OPENED = "calc_doc_opened"
    """Doc Opened Spreadsheet see :py:meth:`Calc.open_doc() <.office.calc.Calc.open_doc>`"""

    DOC_SS = "calc_doc_ss"
    """Doc Get Spreadsheet see :py:meth:`Calc.get_ss_doc() <.office.calc.Calc.get_ss_doc>`"""

    DOC_CREATING = "calc_doc_creating"
    """Doc Creating see :py:meth:`Calc.create_doc() <.office.calc.Calc.create_doc>`"""
    DOC_CREATED = "calc_doc_created"
    """Doc Created see :py:meth:`Calc.create_doc() <.office.calc.Calc.create_doc>`"""

    DOC_SAVING = "calc_doc_saving"
    """Doc Saving Spreadsheet document see :py:meth:`Calc.save_doc() <.office.calc.Calc.save_doc>`"""
    DOC_SAVED = "calc_doc_saved"
    """Doc Saving Spreadsheet document see :py:meth:`Calc.save_doc() <.office.calc.Calc.save_doc>`"""

    SHEET_GETTING = "calc_sheet_getting"
    """Sheet Getting see :py:meth:`Calc.get_sheet() <.office.calc.Calc.get_sheet>`"""
    SHEET_GET = "calc_sheet_get"
    """Sheet Get see :py:meth:`Calc.get_sheet() <.office.calc.Calc.get_sheet>`"""

    SHEET_INSERTING = "calc_sheet_inserting"
    """Sheet Inserting see :py:meth:`Calc.insert_sheet() <.office.calc.Calc.insert_sheet>`"""
    SHEET_INSERTED = "calc_sheet_inserted:"
    """Sheet Inserted see :py:meth:`Calc.insert_sheet() <.office.calc.Calc.insert_sheet>`"""

    SHEET_REMOVING = "calc_sheet_removing"
    """Sheet Removing see :py:meth:`Calc.remove_sheet() <.office.calc.Calc.remove_sheet>`"""
    SHEET_REMOVED = "calc_sheet_removed"
    """Sheet Removed see :py:meth:`Calc.remove_sheet() <.office.calc.Calc.remove_sheet>`"""

    SHEET_MOVING = "calc_sheet_moving"
    """Sheet Moving see :py:meth:`Calc.move_sheet() <.office.calc.Calc.move_sheet>`"""
    SHEET_MOVED = "calc_sheet_moved"
    """Sheet Moved see :py:meth:`Calc.move_sheet() <.office.calc.Calc.move_sheet>`"""

    SHEET_ACTIVATING = "calc_sheet_activating"
    """Sheet Activating see :py:meth:`Calc.set_active_sheet() <.office.calc.Calc.set_active_sheet>`"""
    SHEET_ACTIVATED = "calc_sheet_activated"
    """Sheet Activated see :py:meth:`Calc.set_active_sheet() <.office.calc.Calc.set_active_sheet>`"""

    SHEET_ROW_INSERTING = "calc_sheet_row_inserting"
    """Sheet row inserting see :py:meth:`Calc.insert_row() <.office.calc.Calc.insert_row>`"""
    SHEET_ROW_INSERTED = "calc_sheet_row_inserted"
    """Sheet row inserted see :py:meth:`Calc.insert_row() <.office.calc.Calc.insert_row>`"""

    SHEET_ROW_DELETING = "calc_sheet_row_deleting"
    """Sheet row deleting see :py:meth:`Calc.delete_row() <.office.calc.Calc.delete_row>`"""
    SHEET_ROW_DELETED = "calc_sheet_row_deleted"
    """Sheet row deleted see :py:meth:`Calc.delete_row() <.office.calc.Calc.delete_row>`"""

    SHEET_COL_INSERTING = "calc_sheet_col_inserting"
    """Sheet column inserting see :py:meth:`Calc.insert_column() <.office.calc.Calc.insert_column>`"""
    SHEET_COL_INSERTED = "calc_sheet_col_inserted"
    """Sheet column inserted see :py:meth:`Calc.insert_column() <.office.calc.Calc.insert_column>`"""

    SHEET_COL_DELETING = "calc_sheet_col_deleting"
    """Sheet column deleting see :py:meth:`Calc.delete_column() <.office.calc.Calc.delete_column>`"""
    SHEET_COL_DELETED = "calc_sheet_col_deleted"
    """Sheet column deleted see :py:meth:`Calc.delete_column() <.office.calc.Calc.delete_column>`"""

    SHEET_COL_WIDTH_SETTING = "calc_sheet_col_width_setting"
    """Sheet setting column width see :py:meth:`Calc.set_col_width() <.office.calc.Calc.set_col_width>`"""
    SHEET_COL_WIDTH_SET = "calc_sheet_col_width_set"
    """Sheet set column width see :py:meth:`Calc.set_col_width() <.office.calc.Calc.set_col_width>`"""

    SHEET_ROW_HEIGHT_SETTING = "calc_sheet_row_height_setting"
    """Sheet setting row height see :py:meth:`Calc.set_row_height() <.office.calc.Calc.set_row_height>`"""
    SHEET_ROW_HEIGHT_SET = "calc_sheet_row_height_set"
    """Sheet set row height see :py:meth:`Calc.set_row_height() <.office.calc.Calc.set_row_height>`"""

    CELLS_INSERTING = "calc_cells_inserting"
    """Cells Inserting see :py:meth:`Calc.insert_cells() <.office.calc.Calc.insert_cells>`"""
    CELLS_INSERTED = "calc_cells_inserted"
    """Cells Inserted see :py:meth:`Calc.insert_cells() <.office.calc.Calc.insert_cells>`"""

    CELLS_DELETING = "calc_cells_deleting"
    """Cells Deleting see :py:meth:`Calc.delete_cells() <.office.calc.Calc.delete_cells>`"""
    CELLS_DELETED = "calc_cells_deleted"
    """Cells Deleted see :py:meth:`Calc.delete_cells() <.office.calc.Calc.delete_cells>`"""

    CELLS_CLEARING = "calc_cells_clearing"
    """Cells Clearing see :py:meth:`Calc.clear_cells() <.office.calc.Calc.clear_cells>`"""
    CELLS_CLEARED = "calc_cells_cleared"
    """Cells Cleared see :py:meth:`Calc.clear_cells() <.office.calc.Calc.clear_cells>`"""

    CELLS_BORDER_REMOVING = "calc_cells_border_removing"
    """Cells Border Removing see :py:meth:`Calc.remove_border() <.office.calc.Calc.remove_border>`"""
    CELLS_BORDER_REMOVED = "calc_cells_border_removed"
    """Cells Border Removed see :py:meth:`Calc.remove_border() <.office.calc.Calc.remove_border>`"""

    CELLS_BORDER_ADDING = "calc_cells_border_adding"
    """Cells Border Adding see :py:meth:`Calc.add_border() <.office.calc.Calc.add_border>`"""
    CELLS_BORDER_ADDED = "calc_cells_border_added"
    """Cells Border Added see :py:meth:`Calc.add_border() <.office.calc.Calc.add_border>`"""

    CELLS_HIGH_LIGHTING = "calc_cells_high_lighting"
    """Cells Highlighting see :py:meth:`Calc.highlight_range() <.office.calc.Calc.highlight_range>`"""
    CELLS_HIGH_LIGHTED = "calc_cells_high_lighted"
    """Cells Highlighted see :py:meth:`Calc.highlight_range() <.office.calc.Calc.highlight_range>`"""



# coding: utf-8
from __future__ import annotations
import os
from typing import Any
import uno

from ..mock import mock_g

if mock_g.DOCS_BUILDING:

    # mock class
    from sphinx.ext.autodoc.mock import _MockObject
    class UnoEnum(_MockObject):
        pass
    
    # alternative below that includes a little more informaton.
    class _UnoEnum:
        # Note the __sphinx_mock__ attrib
        __sphinx_mock__ = True
        _loaded = {}
        _initialized = False
        def __new__(cls, type_name: str):

            if type_name in cls._loaded:
                return cls._loaded[type_name]
            e = super().__new__(cls)
            e.__dict__["_type_name"] = type_name
            cls._loaded[type_name] = e
            return e
            
        def __init__(self, type_name: str) -> None:
            self._initialized = True
        
        def __getattr__(self, __name: str) -> uno.Enum | Any:
            if self._initialized:
                # Provide the caller attributes in whatever ways interest you.
                try:
                    key = __name
                    e = f"<Enum instance {self._type_name} ({__name})>"
                    self.__dict__[key] = e
                    return self.__dict__[key]
                except Exception:
                    raise AttributeError(f"Enum {self._type_name} has no attribute {__name}")
            else:
                try:
                    return self.__dict__[__name] # Transparent access to instance vars.
                except KeyError:
                    raise AttributeError(__name)

        def __setattr__(self, key, value):
            if self._initialized:
                pass # Provide caller ways to set attributes in whatever ways.
            else:
                self.__dict__[key] = value # Transparent access.

        @property
        def typeName(self) -> str:
            """
            Gets Uno Type Name

            Returns:
                str: Uno type Name
            """
            return self._type_name
else:
    class UnoEnum:
        """
        Get access to Uno Enum values without having to directly import them.
        
        This is a singleton class.
        Repeated call to constructor for any given Enum type,
        such as ``com.sun.star.sheet.FillMode`` will return the same object
        
        Example:
        
            .. code-block:: python
            
                FillMode = UnoEnum("com.sun.star.sheet.FillMode")
                assert FillMode.LINEAR.value == "LINEAR"
                
                MyEnum = UnoEnum("com.sun.star.sheet.FillMode")
                assert FillMode is MyEnum # singleton, same instances
        
        Note:
            Uno enums can not be imported directly in python.

            ``from com.sun.star.sheet import FillMode`` is not a valid import
            and results in an import error.
            
            ``from com.sun.star.sheet.FillMode import LINEAR, SIMPLE, GROWTH`` is valid import but cumbersome.
            
            Best of both worlds. using ``typing.TYPE_CHECKING``.
            ``typing.TYPE_CHECKING`` is always false at runtime. This means anything imported
            in a ``typing.TYPE_CHECKING`` block is ignored at runtime.
            
            .. code-block:: python
            
                from typing import cast, TYPE_CHECKING
                if TYPE_CHECKING:
                    # uno enums are not able to be imported at runtime
                    from com.sun.star.sheet.FillMode as UnoFillMode
                
                FillMode = cast("UnoFillMode", UnoEnum("com.sun.star.sheet.FillMode"))
            
            In the above example ``FillMode`` will have full typing (intellsense) at design time.
            At runtime cast simply returns the UnoEnum instance.

            Note that ``"UnoFillMode"`` is wrapped in quotes. This is necessary for typing reasons.
            Without wrapping in quotes python will look for the import at runtime
            which will not be available becuase is in a ``TYPE_CHECKING`` block.

        See Also:
            :py:class:`~.uno_const.UnoConst`
        """
        _loaded = {}
        _initialized = False  # This class var is important. It is always False.
                            # The instances will override this with their own, 
                            # set to True.
        def __new__(cls, type_name: str):
            # if _DOCS_BUILDING:
            #     # provision for sphinx autodoc
            #     # sphinx autodoc errors if this class is called
            #     # such as when a class instance is assigned to a constant
            #     # see Write class
            #     return type_name
            if type_name in cls._loaded:
                return cls._loaded[type_name]
            e = super().__new__(cls)
            e.__dict__["_type_name"] = type_name
            cls._loaded[type_name] = e
            return e
            
        def __init__(self, type_name: str) -> None:
            """
            Constructor
            
            Get access to Uno Enum values without having to directly import them.

            Args:
                type_name (str): The namespace of the enum as a string.
            """
            self._initialized = True
        
        def __getattr__(self, __name: str) -> uno.Enum | Any:
            if self._initialized:
                # Provide the caller attributes in whatever ways interest you.
                try:
                    key = __name
                    e = uno.Enum(self._type_name, __name)
                    self.__dict__[key] = e
                    return self.__dict__[key]
                except Exception:
                    raise AttributeError(f"Enum {self._type_name} has no attribute {__name}")
            else:
                try:
                    return self.__dict__[__name] # Transparent access to instance vars.
                except KeyError:
                    raise AttributeError(__name)

        def __setattr__(self, key, value):
            if self._initialized:
                pass # Provide caller ways to set attributes in whatever ways.
            else:
                self.__dict__[key] = value # Transparent access.

        @property
        def typeName(self) -> str:
            """
            Gets Uno Type Name

            Returns:
                str: Uno type Name
            """
            return self._type_name

    __all__ = ("UnoEnum",)


from enum import Enum
from . import kind_helper


class DrawingBitmapKind(str, Enum):
    """
    Drawing (some) Bitmap Values.
    The actual value may change from version to version and may be different
    in other languages.

    These are just quick suggestion that show up in Draw Bitmap options.
    """

    BATHROOM_TILES = "Bathroom Tiles"
    BRICK_WALL = "Brick Wall"
    CARDBOARD = "Cardboard"
    COFFEE_BEANS = "Coffee Beans"
    COLOR_STRIPES = "Color Stripes"
    COLORFUL_PEBBLES = "Colorful Pebbles"
    CONCRETE = "Concrete"
    FENCE = "Fence"
    FLORAL = "Floral"
    ICE_LIGHT = "Ice light"
    INVOICE_PAPER = "Invoice Paper"
    LAWN = "Lawn"
    LITTLE_CLOUDS = "Little Clouds"
    MAPLE_LEAVES = "Maple Leaves"
    MARBLE = "Marble"
    PAINTED_WHITE = "Painted White"
    PAPER_CRUMPLED = "Paper Crumpled"
    PAPER_GRAPH = "Paper Graph"
    PAPER_TEXTURE = "Paper Texture"
    PARCHMENT_PAPER = "Parchment Paper"
    POOL = "Pool"
    SAND_LIGHT = "Sand light"
    SPACE = "Space"
    STONE_WALL = "Stone Wall"
    STONE = "Stone"
    STUDIO = "Studio"
    SURFACE = "Surface"
    WALL_OF_ROCK = "Wall of Rock"
    WHITE_DIFFUSION = "White Diffusion"
    ZEBRA = "Zebra"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "DrawingBitmapKind":
        """
        Gets an ``DrawingBitmapKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``DrawingBitmapKind`` instance.

        Returns:
            DrawingBitmapKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, DrawingBitmapKind)



from enum import Enum
from . import kind_helper


class DrawingLayerKind(str, Enum):
    """
    Drawing Layer Values.

    See Also:
        :py:meth:`~.Draw.get_layer`
    """

    BACK_GROUND = "background"
    BACK_GROUND_OBJECTS = "backgroundobjects"
    CONTROLS = "controls"
    LAYOUT = "layout"
    MEASURE_LINES = "measurelines"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "DrawingLayerKind":
        """
        Gets an ``DrawingLayerKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``DrawingLayerKind`` instance.

        Returns:
            DrawingLayerKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, DrawingLayerKind)



# coding: utf-8
from __future__ import annotations
from typing import Callable, Sequence, TypeVar, Union, Any, Tuple, List, Dict
from os import PathLike

import uno
from com.sun.star.text import XText
from com.sun.star.text import XTextCursor
from com.sun.star.text import XTextDocument

from ..events.args import event_args

PathOrStr = Union[str, PathLike]
"""Path like object or string"""

UnoInterface = object
"""Represents a uno interface class. Any uno Class that starts with X"""

T = TypeVar("T")

Row = Sequence[Any]
"""Represents a Row of a Table."""

DictRow = Dict[str, Any]
"""
Represents a Row of a Table as a dictionary of Key, value.
Where the key is the column name.
"""

DictTable = Sequence[DictRow]
"""
Represents a Dictionary table.
Each Element in the sequence as DictRow (dictionary of key, value)
with key as column name.
"""

Column = Sequence[Any]
"""Represents a Column of a Table."""

Table = Sequence[Row]
"""Represents a 2-D Table of Rows and Columns"""

TupleArray = Tuple[Tuple[Any, ...], ...]
"""Table like tuples with rows and columns"""

FloatList = List[float]
"""List of Floats"""

FloatTable = List[FloatList]
"""Table like array of floats with rows and columns"""

DocOrCursor = Union[XTextDocument, XTextCursor]
"""Type of Text Document or Cursor"""

DocOrText = Union[XTextDocument, XText]
"""Type of Text Document of Text"""


EventCallback = Callable[[object, event_args.EventArgs], None]
"""Event Callback"""


# coding: utf-8
"""
Calc Named Events.
"""
from __future__ import annotations
from typing import NamedTuple


class WriteNamedEvent(NamedTuple):
    """
    Named events for :py:class:`~.office.wite.Write` class
    """

    DOC_OPENING = "write_doc_opening"
    """Doc Opening Write document see :py:meth:`Write.open_doc() <.office.write.Write.open_doc>`"""
    DOC_OPENED = "write_doc_opened"
    """Doc Opened Write document see :py:meth:`Write.open_doc() <.office.write.Write.open_doc>`"""

    DOC_CREATING = "write_doc_creating"
    """Doc Creating Write document see :py:meth:`Write.create_doc() <.office.write.Write.create_doc>`"""
    DOC_CREATED = "write_doc_created"
    """Doc Created Write document see :py:meth:`Write.create_doc() <.office.write.Write.create_doc>`"""

    DOC_CLOSING = "write_doc_closing"
    """Doc Closing Write document see :py:meth:`Write.close_doc() <.office.write.Write.close_doc>`"""
    DOC_CLOSED = "write_doc_closed"
    """Doc Closed Write document see :py:meth:`Write.close_doc() <.office.write.Write.close_doc>`"""

    DOC_SAVING = "write_doc_saving"
    """Doc Closing Write document see :py:meth:`Write.save_doc() <.office.write.Write.save_doc>`"""
    DOC_SAVED = "write_doc_saved"
    """Doc Closed Write document see :py:meth:`Write.save_doc() <.office.write.Write.save_doc>`"""

    DOC_TMPL_CREATING = "write_doc_tmpl_creating"
    """Doc Creating Write document see :py:meth:`Write.create_doc_from_template() <.office.write.Write.create_doc_from_template>`"""
    DOC_TMPL_CREATED = "write_doc_tmpl_created"
    """Doc Created Write document see :py:meth:`Write.create_doc_from_template() <.office.write.Write.create_doc_from_template>`"""

    WORD_SELECTING = "write_word_selecting"
    """Word Selecting see :py:meth:`Write.select_next_word() <.office.write.Write.select_next_word>`"""
    WORD_SELECTED = "write_word_selected"
    """Word Selected see :py:meth:`Write.select_next_word() <.office.write.Write.select_next_word>`"""

    DOC_TEXT = "write_text_doc"
    """Got a Write document see :py:meth:`Write.get_text_doc() <.office.write.Write.get_text_doc>`"""

    PAGE_FORRMAT_SETTING = "write_page_format_setting"
    """Page Format Setting see :py:meth:`Write.set_page_format() <.office.write.Write.set_page_format>`"""
    PAGE_FORRMAT_SET = "write_page_format_setting"
    """Page Format Set see :py:meth:`Write.set_page_format() <.office.write.Write.set_page_format>`"""

    FORMULA_ADDING = "write_forumla_adding"
    """Formula Adding see :py:meth:`Write.add_formula() <.office.write.Write.add_formula>`"""
    FORMULA_ADDED = "write_forumla_added"
    """Formula Added see :py:meth:`Write.add_formula() <.office.write.Write.add_formula>`"""

    HYPER_LINK_ADDING = "write_hyper_link_adding"
    """Hyperlink Adding see :py:meth:`Write.add_hyperlink() <.office.write.Write.add_hyperlink>`"""
    HYPER_LINK_ADDED = "write_hyper_link_added"
    """Hyperlink Added see :py:meth:`Write.add_hyperlink() <.office.write.Write.add_hyperlink>`"""

    BOOKMARK_ADDING = "write_bookmark_adding"
    """Bookmark Adding see :py:meth:`Write.add_bookmark() <.office.write.Write.add_bookmark>`"""
    BOOKMARK_ADDIED = "write_bookmark_added"
    """Bookmark Added see :py:meth:`Write.add_bookmark() <.office.write.Write.add_bookmark>`"""

    TEXT_FRAME_ADDING = "write_text_frame_adding"
    """Text frame Adding see :py:meth:`Write.add_text_frame() <.office.write.Write.add_text_frame>`"""
    TEXT_FRAME_ADDED = "write_text_frame_added"
    """Text frame Added see :py:meth:`Write.add_text_frame() <.office.write.Write.add_text_frame>`"""

    TABLE_ADDING = "write_tabel_adding"
    """Table Adding see :py:meth:`Write.add_table() <.office.write.Write.add_table>`"""
    TABLE_ADDED = "write_tabel_added"
    """Table Added see :py:meth:`Write.add_table() <.office.write.Write.add_table>`"""

    IMAGE_LNIK_ADDING = "write_image_link_adding"
    """Image Link Adding see :py:meth:`Write.add_image_link() <.office.write.Write.add_image_link>`"""
    IMAGE_LNIK_ADDED = "write_image_link_added"
    """Image Link Added see :py:meth:`Write.add_image_link() <.office.write.Write.add_image_link>`"""

    IMAGE_SHAPE_ADDING = "write_image_shape_adding"
    """Image Shape Adding see :py:meth:`Write.add_image_shape() <.office.write.Write.add_image_shape>`"""
    IMAGE_SHAPE_ADDED = "write_image_shape_added"
    """Image Shape Added see :py:meth:`Write.add_image_shape() <.office.write.Write.add_image_shape>`"""

    CONFIGURED_SERVICES_SETTING = "write_configure_services_setting"
    """Services configuration setting see :py:meth:`Write.set_configured_services() <.office.write.Write.set_configured_services>`"""
    CONFIGURED_SERVICES_SET = "write_configure_services_set"
    """Services configuration set see :py:meth:`Write.set_configured_services() <.office.write.Write.set_configured_services>`"""



from .kind_base import KindBase
from . import kind_helper


class FormComponentKind(KindBase):
    """
    Values used with ``com.sun.star.form.component.*``

    See Also:
        `component API <https://api.libreoffice.org/docs/idl/ref/namespacecom_1_1sun_1_1star_1_1form_1_1component.html>`_
    """

    CHECK_BOX = "CheckBox"
    COMBO_BOX = "ComboBox"
    COMMAND_BUTTON = "CommandButton"
    CURRENCY_FIELD = "CurrencyField"
    DATABASE_CHECK_BOX = "DatabaseCheckBox"
    DATABASE_COMBO_BOX = "DatabaseComboBox"
    DATABASE_CURRENCY_FIELD = "DatabaseCurrencyField"
    DATABASE_DATE_FIELD = "DatabaseDateField"
    DATABASE_FORMATTED_FIELD = "DatabaseFormattedField"
    DATABASE_IMAGE_CONTROL = "DatabaseImageControl"
    DATABASE_LIST_BOX = "DatabaseListBox"
    DATABASE_NUMERIC_FIELD = "DatabaseNumericField"
    DATABASE_PATTERN_FIELD = "DatabasePatternField"
    DATABASE_RADIO_BUTTON = "DatabaseRadioButton"
    DATABASE_TEXT_FIELD = "DatabaseTextField"
    DATABASE_TIME_FIELD = "DatabaseTimeField"
    DATE_FIELD = "DateField"
    FILE_CONTROL = "FileControl"
    FIXED_TEXT = "FixedText"
    FORMATTED_FIELD = "FormattedField"
    GRID_CONTROL = "GridControl"
    GROUP_BOX = "GroupBox"
    HIDDEN_CONTROL = "HiddenControl"
    HTML_FORM = "HTMLForm"
    IMAGE_BUTTON = "ImageButton"
    LIST_BOX = "ListBox"
    NAVIGATION_TOOL_BAR = "NavigationToolBar"
    NUMERIC_FIELD = "NumericField"
    PATTERN_FIELD = "PatternField"
    RADIO_BUTTON = "RadioButton"
    RICH_TEXT_CONTROL = "RichTextControl"
    SCROLL_BAR = "ScrollBar"
    SPIN_BUTTON = "SpinButton"
    SUBMIT_BUTTON = "SubmitButton"
    TEXT_FIELD = "TextField"
    TIME_FIELD = "TimeField"

    def to_namespace(self) -> str:
        """Gets full name-space value of instance"""
        return f"com.sun.star.form.component.{self.value}"

    @staticmethod
    def from_str(s: str) -> "FormComponentKind":
        """
        Gets an ``FormComponentKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``FormComponentKind`` instance.

        Returns:
            FormComponentKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, FormComponentKind)



from .kind_base import KindBase
from . import kind_helper


class PresentationKind(KindBase):
    """
    Values used with ``com.sun.star.presentation.*``

    See Also:
        `Presentation API <https://api.libreoffice.org/docs/idl/ref/namespacecom_1_1sun_1_1star_1_1presentation.html>`_
    """

    CHART_SHAPE = "ChartShape"
    CUSTOM_PRESENTATION = "CustomPresentation"
    CUSTOM_PRESENTATION_ACCESS = "CustomPresentationAccess"
    DATE_TIME_SHAPE = "DateTimeShape"
    DOCUMENT_SETTINGS = "DocumentSettings"
    DRAW_PAGE = "DrawPage"
    FOOTER_SHAPE = "FooterShape"
    GRAPHIC_OBJECT_SHAPE = "GraphicObjectShape"
    HANDOUT_SHAPE = "HandoutShape"
    HANDOUT_VIEW = "HandoutView"
    HEADER_SHAPE = "HeaderShape"
    NOTES_SHAPE = "NotesShape"
    NOTES_VIEW = "NotesView"
    OLE2_SHAPE = "OLE2Shape"
    OUTLINE_VIEW = "OutlineView"
    OUTLINER_SHAPE = "OutlinerShape"
    PAGE_SHAPE = "PageShape"
    PARAGRAPH_TARGET = "ParagraphTarget"
    PRESENTATION = "Presentation"
    PRESENTATION_DOCUMENT = "PresentationDocument"
    PRESENTATION_VIEW = "PresentationView"
    PRESENTATION2 = "Presentation2"
    PREVIEW_VIEW = "PreviewView"
    SHAPE = "Shape"
    SLIDE_NUMBER_SHAPE = "SlideNumberShape"
    SLIDE_SHOW = "SlideShow"
    SLIDES_VIEW = "SlidesView"
    SUBTITLE_SHAPE = "SubtitleShape"
    TITLE_TEXT_SHAPE = "TitleTextShape"
    TRANSITION_FACTORY = "TransitionFactory"

    def to_namespace(self) -> str:
        """Gets full name-space value of instance"""
        return f"com.sun.star.presentation.{self.value}"

    @staticmethod
    def from_str(s: str) -> "PresentationKind":
        """
        Gets an ``PresentationKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``PresentationKind`` instance.

        Returns:
            PresentationKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, PresentationKind)



# coding: utf-8
from __future__ import annotations
from typing import Union
import os
import sys
import shutil
import __main__
from pathlib import Path
from typing import overload
from .sys_info import SysInfo

# do not import from type_var here.
# this module is used by uno_lnk.py and type_var imports uno
# PathOrStr = Union[str, os.PathLike]

# python path on mac:  /Applications/LibreOffice.app/Contents/Resources/python
#   https://ask.libreoffice.org/t/where-is-the-python-executable-embedded-in-libreoffice-on-macos/50042


PLATFORM = SysInfo.get_platform()
if PLATFORM == SysInfo.PlatformEnum.WINDOWS:
    import winreg

_INSTALL_PATH = None


def get_soffice_install_path() -> Path:
    """
    Gets the Soffice install path.

    For windows this will be something like: ``C:\\Program Files\\LibreOffice``.
    For Linux this will be something like: ``/usr/lib/libreoffice``

    Returns:
        Path: install as Path.
    """
    global _INSTALL_PATH
    if _INSTALL_PATH is not None:
        return _INSTALL_PATH
    if PLATFORM == SysInfo.PlatformEnum.WINDOWS:
        # get the path location from registery
        value = ""
        for _key in (
            # LibreOffice 3.4.5,6,7 on Windows
            "SOFTWARE\\LibreOffice\\UNO\\InstallPath",
            # OpenOffice 3.3
            "SOFTWARE\\OpenOffice.org\\UNO\\InstallPath",
        ):
            try:
                value = winreg.QueryValue(winreg.HKEY_LOCAL_MACHINE, _key)
            except Exception as detail:
                value = ""
                _errMess = "%s" % detail
            else:
                break  # first existing key will do
        if value != "":
            _INSTALL_PATH = Path("\\".join(value.split("\\")[:-1]))  # drop the program
            return _INSTALL_PATH

        # failed to get path from registery. Going Manual
        soffice = "soffice.exe"
        p_sf = Path(os.environ["PROGRAMFILES"], "LibreOffice", "program", soffice)
        if p_sf.exists() is False or p_sf.is_file() is False:
            p_sf = Path(os.environ["PROGRAMFILES(X86)"], "LibreOffice", "program", soffice)
        if not p_sf.exists():
            raise FileNotFoundError(f"LibreOffice '{p_sf}' not found.")
        if not p_sf.is_file():
            raise IsADirectoryError(f"LibreOffice '{p_sf}' is not a file.")
        # drop \program\soffice.exe
        # expect C:\Program Files\LibreOffice
        _INSTALL_PATH = p_sf.parent.parent
        return _INSTALL_PATH

    elif PLATFORM == SysInfo.PlatformEnum.MAC:
        _INSTALL_PATH = Path("/Applications/LibreOffice.app/Contents/MacOS")
        return _INSTALL_PATH
    else:
        # unix
        soffice = "soffice"
        # search system path
        s = shutil.which(soffice)
        p_sf = None
        if s is not None:
            # expect '/usr/bin/soffice'
            if os.path.islink(s):
                p_sf = Path(os.path.realpath(s)).parent
            else:
                p_sf = Path(s).parent
        if p_sf is None:
            s = "/usr/bin/soffice"
            if os.path.islink(s):
                p_sf = Path(os.path.realpath(s)).parent
            else:
                p_sf = Path(s).parent
        if not p_sf.exists():
            raise FileNotFoundError(f"LibreOffice '{p_sf}' not found.")
        if not p_sf.is_file():
            raise IsADirectoryError(f"LibreOffice '{p_sf}' is not a file.")
        # drop /program/soffice
        _INSTALL_PATH = p_sf.parent.parent
        return _INSTALL_PATH


def get_soffice_path() -> Path:
    """
    Gets path to soffice

    Returns:
        Path: path to soffice
    """
    if PLATFORM == SysInfo.PlatformEnum.WINDOWS:
        return Path(get_lo_path(), "soffice.exe")
    return Path(get_lo_path(), "soffice")


def get_uno_path() -> Path:
    """
    Searches known paths for path that contains uno.py

    This path is different for windows and linux.
    Typically for Windows ``C:\\Program Files\\LibreOffice\\program``
    Typically for Linux ``/usr/lib/python3/dist-packages``

    Raises:
        FileNotFoundError: if path is not found
        NotADirectoryError: if path is not a directory

    Returns:
        Path: First found path.
    """
    if PLATFORM == SysInfo.PlatformEnum.WINDOWS:

        p_uno = Path(os.environ["PROGRAMFILES"], "LibreOffice", "program")
        if p_uno.exists() is False or p_uno.is_dir() is False:
            p_uno = Path(os.environ["PROGRAMFILES(X86)"], "LibreOffice", "program")
        if not p_uno.exists():
            raise FileNotFoundError("Uno Source Dir not found.")
        if not p_uno.is_dir():
            raise NotADirectoryError("UNO source is not a Directory")
        return p_uno
    elif PLATFORM == SysInfo.PlatformEnum.MAC:
        return Path("/Applications/LibreOffice.app/Contents/MacOS/soffice")
    else:
        p_uno = Path("/usr/lib/python3/dist-packages")
        if not p_uno.exists():
            raise FileNotFoundError("Uno Source Dir not found.")
        if not p_uno.is_dir():
            raise NotADirectoryError("UNO source is not a Directory")
        return p_uno


def get_lo_path() -> Path:
    """
    Searches known paths for path that contains LibreOffice ``soffice``.

    This path is different for windows and linux.
    Typically for Windows ``C:\\Program Files\\LibreOffice\\program``
    Typically for Linux ``/usr/bin/soffice``

    Raises:
        FileNotFoundError: if path is not found
        NotADirectoryError: if path is not a directory

    Returns:
        Path: First found path.
    """
    if PLATFORM == SysInfo.PlatformEnum.WINDOWS:
        return Path(get_soffice_install_path(), "program")

    elif PLATFORM == SysInfo.PlatformEnum.MAC:
        return Path("/Applications/LibreOffice.app/Contents/MacOS")
    else:
        # search system path
        s = shutil.which("soffice")
        p_sf = None
        if s is not None:
            # expect '/usr/bin/soffice'
            if os.path.islink(s):
                # follow link
                p_sf = Path(os.path.realpath(s)).parent
            else:
                p_sf = Path(s).parent
        if p_sf is None:
            p_sf = Path("/usr/bin/soffice")
            if p_sf.exists() is False or p_sf.is_file() is False:
                raise FileNotFoundError("LibreOffice Source Dir not found.")
            p_sf = p_sf.parent

        if not p_sf.exists():
            raise FileNotFoundError("LibreOffice Source Dir not found.")
        if not p_sf.is_dir():
            raise NotADirectoryError("LibreOffice source is not a Directory")
        return p_sf


def get_lo_python_ex() -> str:
    """
    Gets the python executable for different environments.

    In Linux this is the current python executable.
    If a virtual environment is activated then that will be the
    python executable that is returned.

    In Windows this is the ``python.exe`` file in LibreOffice.
    Typically for Windows ``C:\\Program Files\\LibreOffice\\program\\python.exe``

    Raises:
        FileNotFoundError: In Windows if ``python.exe`` is not found.
        NotADirectoryError: In Windows if ``python.exe`` is not a file.

    Returns:
        str: file location of python executable.
    """
    if PLATFORM == SysInfo.PlatformEnum.WINDOWS:
        p = Path(get_lo_path(), "python.exe")

        if not p.exists():
            raise FileNotFoundError("LibreOffice python executable not found.")
        if not p.is_file():
            raise NotADirectoryError("LibreOffice  python executable is not a file")
        return str(p)
    else:
        return sys.executable


@overload
def mkdirp(dest_dir: str) -> None:
    ...


@overload
def mkdirp(dest_dir: Path) -> None:
    ...


def mkdirp(dest_dir: Union[str, os.PathLike]) -> None:
    """
    Creates path and subpaths not existing.

    Args:
        dest_dir (str | PathLike]): PathLike object
    """
    # Python ≥ 3.5
    pth = Path(dest_dir)
    if not pth.is_absolute():
        pth = pth.resolve()
    pth.mkdir(parents=True, exist_ok=True)



class DrawViewDispatch:
    """
    Draw View Dispatch Commands

    See Also:
        - :py:meth:`.Lo.dispatch_cmd`
        - `Draw Dispatch commands <https://wiki.documentfoundation.org/Development/DispatchCommands#Draw_.2F_Impress>`_
    """

    CLEAR_UNDO_STACK = "ClearUndoStack"
    COLOR_VIEW = "ColorView"
    """Black & White View"""
    DIA_MODE = "DiaMode"
    """Slide Sorter"""
    DISPLAY_MASTER_BACKGROUND = "DisplayMasterBackground"
    """Master Background"""
    GRID_FRONT = "GridFront"
    """Grid to Front"""
    HELPLINES_FRONT = "HelplinesFront"
    """Snap Guides to Front"""
    HELPLINES_USE = "HelplinesUse"
    """Snap to Snap Guides"""
    HELPLINES_VISIBLE = "HelplinesVisible"
    """Display Snap Guides"""
    LAYER_MODE = "LayerMode"
    """
    Layer
    
    Args
        - IsActive (bool)
        - WhatLayer (int)
    """
    LAYOUT_STATUS = "LayoutStatus"
    """Layout"""
    MASTER_PAGE = "MasterPage"
    """
    Master
    
    Args
        - IsActive (bool)
    """
    NEXT_ANNOTATION = "NextAnnotation"
    NOTES_CHILD_WINDOW = "NotesChildWindow"
    NOTES_MASTER_PAGE = "NotesMasterPage"
    """Master Notes"""
    NOTES_MODE = "NotesMode"
    """Notes"""
    OBJECT_POSITION = "ObjectPosition"
    """Arrange"""
    OUTLINE_MODE = "OutlineMode"
    """Outline"""
    OUTPUT_QUALITY_BLACK_WHITE = "OutputQualityBlackWhite"
    """Black and White"""
    OUTPUT_QUALITY_COLOR = "OutputQualityColor"
    """Color"""
    OUTPUT_QUALITY_CONTRAST = "OutputQualityContrast"
    """High Contrast"""
    OUTPUT_QUALITY_GRAYSCALE = "OutputQualityGrayscale"
    """Gray scale"""
    PAGE_MODE = "PageMode"
    """
    Normal
    
    Args
        - IsActive (bool)
        - WhatKind (int)
    """
    PAGE_STATUS = "PageStatus"
    """Slide/Layer"""
    PRESENTATION = "Presentation"
    """Start from First Slide"""
    PRESENTATION_CURRENT_SLIDE = "PresentationCurrentSlide"
    """Start from Current Slide"""
    PRESENTATION_END = "PresentationEnd"
    PREVIEW_STATE = "PreviewState"
    PREVIOUS_ANNOTATION = "PreviousAnnotation"
    SCALE = "Scale"
    SLIDE_MASTER_PAGE = "SlideMasterPage"
    """Master Slide"""
    SNAP_BORDER = "SnapBorder"
    """Snap to Page Margins"""
    SNAP_FRAME = "SnapFrame"
    SNAP_POINTS = "SnapPoints"
    """Snap to Object Points"""
    SOLID_CREATE = "SolidCreate"
    """Modify Object with Attributes"""
    SWITCH_LAYER = "SwitchLayer"
    """
    Args
        - WhatLayer (int)
    """
    SWITCH_PAGE = "SwitchPage"
    """
    Args
        - WhatPage (int)
        - WhatKind (int)
    """
    TOGGLE_TAB_BAR_VISIBILITY = "ToggleTabBarVisibility"
    """Toggle Views Tab Bar Visibility"""
    ZOOM_MODE = "ZoomMode"
    """Zoom & Pan (Control to Zoom Out, Shift to Pan)"""
    ZOOM_PANNING = "ZoomPanning"
    """Shift"""



from enum import Enum
from . import kind_helper


class LineStyleNameKind(str, Enum):
    """
    Line Style Name Kind.

    Used to encapsulate Line Style Names.
    """

    DASHED = "Dashed"
    DASHES_3_DOTS_3 = "3 Dashes 3 Dots"
    DOTS_2_DASH_1 = "2 Dots 1 Dash"
    DOTS_2_DASHES_3 = "2 Dots 3 Dashes"
    FINE_DASHED = "Fine Dashed"
    FINE_DOTTED = "Fine Dotted"
    LINE_STYLE_9 = "Line Style 9"
    LINE_WITH_FINE_DOTS = "Line with Fine Dots"
    ULTRA_FINE_DOTTED = "Ultrafine Dotted"
    ULTRAFINE_DASHED = "Ultrafine Dashed"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "LineStyleNameKind":
        """
        Gets an ``LineStyleNameKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``LineStyleNameKind`` instance.

        Returns:
            LineStyleNameKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, LineStyleNameKind)



# coding: utf-8
# region Imports
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, List, cast, overload

import uno

from . import lo as mLo
from . import info as mInfo
from . import props as mProps
from . import gui as mGui
from .kind.form_component_kind import FormComponentKind

from com.sun.star.awt import XControl
from com.sun.star.awt import XControlModel
from com.sun.star.beans import XPropertySet
from com.sun.star.container import XChild
from com.sun.star.container import XIndexContainer
from com.sun.star.container import XNameAccess
from com.sun.star.container import XNameAccess
from com.sun.star.container import XNameContainer
from com.sun.star.container import XNamed
from com.sun.star.drawing import XControlShape
from com.sun.star.drawing import XDrawPage
from com.sun.star.drawing import XDrawPagesSupplier
from com.sun.star.drawing import XDrawPageSupplier
from com.sun.star.drawing import XShapes
from com.sun.star.form import XForm
from com.sun.star.form import XFormsSupplier
from com.sun.star.form import XGridColumnFactory
from com.sun.star.lang import XComponent
from com.sun.star.lang import XServiceInfo
from com.sun.star.script import XEventAttacherManager

from ooo.dyn.awt.point import Point
from ooo.dyn.awt.size import Size
from ooo.dyn.form.form_component_type import FormComponentType
from ooo.dyn.form.list_source_type import ListSourceType
from ooo.dyn.script.script_event_descriptor import ScriptEventDescriptor
from ooo.dyn.sdb.command_type import CommandType
from ooo.dyn.text.text_content_anchor_type import TextContentAnchorType

if TYPE_CHECKING:
    from com.sun.star.lang import EventObject
# endregion Imports


class Forms:
    # region    access forms in document
    # region        get_forms()
    @overload
    @classmethod
    def get_forms(cls, obj: XComponent) -> XNameContainer:
        ...

    @overload
    @classmethod
    def get_forms(cls, obj: XDrawPage) -> XNameContainer:
        ...

    @classmethod
    def get_forms(cls, obj: XComponent | XDrawPage) -> XNameContainer:
        """
        Gets Forms

        Args:
            obj (XComponent | XDrawPage): component or draw page

        Returns:
            XNameContainer: name container
        """
        if mLo.Lo.is_uno_interfaces(obj, XDrawPage):
            draw_page = obj
        else:
            draw_page = cls.get_draw_page(obj)

        forms_supp = mLo.Lo.qi(XFormsSupplier, draw_page, True)

        return forms_supp.getForms()

    # endregion     get_forms()

    @staticmethod
    def get_draw_page(doc: XComponent) -> XDrawPage:
        """
        Gets draw page

        Args:
            doc (XComponent): Component

        Raises:
            Exception: If unable to get draw page

        Returns:
            XDrawPage: Draw Page
        """
        try:
            xsupp_page = mLo.Lo.qi(XDrawPageSupplier, doc)
            if xsupp_page is not None:
                return xsupp_page.getDrawPage()

            # doc supports multiple DrawPages
            xsupp_pages = mLo.Lo.qi(XDrawPagesSupplier, doc)

            xpages = xsupp_pages.getDrawPages()
            return mLo.Lo.qi(XDrawPage, xpages.getByIndex(0))
        except Exception as e:
            raise Exception(f"Unable to get draw page: {e}") from e

    # region        get_form()
    @overload
    @classmethod
    def get_form(cls, obj: XComponent) -> XNameContainer:
        ...

    @overload
    @classmethod
    def get_form(cls, obj: XComponent, for_name: str) -> XNameContainer:
        ...

    @overload
    @classmethod
    def get_form(cls, obj: XDrawPage) -> XNameContainer:
        ...

    @classmethod
    def get_form(cls, obj: XComponent | XDrawPage, form_name: str | None = None) -> XNameContainer:
        """
        Gets form as name container

        Args:
            obj (XComponent | XDrawPage): Component or draw page
            form_name (str, optional): the name of form to get.

        Raises:
            Exception: If unable to get form

        Returns:
            XNameContainer: Name container
        """
        if form_name is not None:
            # get_form(cls, obj: XComponent, for_name: str)
            try:
                named_forms = cls.get_forms(obj)
                con = cls.get_form_by_name(form_name, named_forms)
                return mLo.Lo.qi(XForm, con, True)
            except Exception as e:
                raise Exception(f"Unabel to get form: {e}") from e
        try:
            if mLo.Lo.is_uno_interfaces(obj, XDrawPage):
                draw_page = obj
            else:
                draw_page = cls.get_draw_page(obj)

            idx_forms = cls.get_indexed_forms(draw_page)
        except Exception as e:
            raise Exception(f"Unabel to get form: {e}") from e

        try:
            return mLo.Lo.qi(XNameContainer, idx_forms.getByIndex(0))
        except Exception as e:
            raise Exception(f"Could not find default form: {e}") from e

    # endregion     get_form()

    @staticmethod
    def get_form_by_name(form_name: str, named_forms: XNameContainer) -> XNameContainer:
        """
        Get a form by name

        Args:
            form_name (str): form name
            named_forms (XNameContainer): name container

        Raises:
            Exception: If not able to find form

        Returns:
            XNameContainer: Name Container
        """
        try:
            return mLo.Lo.qi(XNameContainer, named_forms.getByName(form_name))
        except Exception as e:
            raise Exception(f'Could not find the form "{form_name}"') from e

    @staticmethod
    def get_indexed_forms(draw_page: XDrawPage) -> XIndexContainer:
        """
        Get index forms

        Args:
            draw_page (XDrawPage): Draw page

        Returns:
            XIndexContainer: Index container
        """
        form_supp = mLo.Lo.qi(XFormsSupplier, draw_page)
        return mLo.Lo.qi(XIndexContainer, form_supp.getForms())

    # region        insert_form()
    @classmethod
    def _insert_form_name_comp(cls, doc: XComponent) -> XNameContainer:
        doc_forms = cls.get_forms(doc)
        return cls._insert_form_namecontainer("GridForm", doc_forms)

    @classmethod
    def _insert_form_namecontainer(cls, form_name: str, named_forms: XNameContainer) -> XNameContainer:
        if named_forms.hasByName(form_name):
            mLo.Lo.print(f'"{form_name}" already exists')
            return cls.get_form_by_name(form_name=form_name, named_forms=named_forms)

        try:
            xnamed_forms = mLo.Lo.create_instance_msf(
                XNameContainer, "com.sun.star.form.component.DataForm", raise_err=True
            )
            named_forms.insertByName(form_name, xnamed_forms)
            return xnamed_forms
        except Exception as e:
            raise Exception(f'Could not insert the form "{form_name}": {e}') from e

    @overload
    @classmethod
    def insert_form(cls, doc: XComponent) -> XNameContainer:
        """
        Insert form

        Args:
            form_name (str): Form name
            doc (XComponent): Component

        Returns:
            XNameContainer: Name Container
        """
        ...

    @overload
    @classmethod
    def insert_form(cls, form_name: str, named_forms: XNameContainer) -> XNameContainer:
        """
        Insert form

        Args:
            form_name (str): Form name
            named_forms (XNameContainer): Name Container

        Returns:
            XNameContainer: Name Container
        """
        ...

    @classmethod
    def insert_form(cls, *args, **kwargs) -> XNameContainer:
        """
        Insert form

        Args:
            form_name (str): Form name
            doc (XComponent): Component
            named_forms (XNameContainer): Name Container

        Returns:
            XNameContainer: Name Container
        """
        ordered_keys = (1, 2)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("form_name", "doc", "named_forms")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("insert_form() got an unexpected keyword argument")
            keys = ("doc", "form_name")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            if count == 1:
                return ka
            ka[2] = kwargs.get("named_forms", None)
            return ka

        if not count in (1, 2):
            raise TypeError("insert_form() got an invalid number of arguments")

        kargs = get_kwargs()
        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        if count == 1:
            return cls._insert_form_name_comp(kargs[1])

        return cls._insert_form_namecontainer(form_name=kargs[1], named_forms=kargs[2])
        # endregion     insert_form()

    @classmethod
    def has_form(cls, doc: XComponent, form_name: str) -> bool:
        """
        Gets if component has form by name

        Args:
            doc (XComponent): Component
            form_name (str): Form name

        Returns:
            bool: ``True`` if has form, Otherwise ``False``
        """
        try:
            draw_page = cls.get_draw_page(doc)
        except Exception:
            return False
        try:
            named_forms_container = cls.get_forms(draw_page)
            if named_forms_container is None:
                mLo.Lo.print("No forms found on page")
                return False
            xnamed_forms = mLo.Lo.qi(XNameAccess, named_forms_container, True)
            return xnamed_forms.hasByName(form_name)
        except Exception:
            return False

    @classmethod
    def show_form_names(cls, doc: XComponent) -> None:
        """
        Prints form names to console.

        Args:
            doc (XComponent): Component

        Returns:
            None:
        """
        form_names_con = cls.get_forms(doc)
        form_names = form_names_con.getElementNames()
        print(f"No. of forms found: {len(form_names)}")
        for name in form_names:
            print(f"  {name}")
        print()

    @classmethod
    def list_forms(cls, obj: XComponent | XNameAccess, tab_str: str = "  ") -> None:
        """
        Prints forms information to console.

        Args:
            obj (XComponent | XNameAccess): Component or Name Access
            tab_str (str, optional): tab string

        Returns:
            None:
        """
        if mLo.Lo.is_uno_interfaces(obj, XComponent):
            xcontainer = cls.get_forms(obj)
        else:
            xcontainer = cast(XNameContainer, obj)
        nms = xcontainer.getElementNames()
        for name in nms:
            try:
                serv_info = mLo.Lo.qi(XServiceInfo, xcontainer.getByName(name), True)
                if serv_info.supportsService("com.sun.star.form.FormComponents"):
                    # this means that the form has been found
                    if mInfo.Info.support_service(serv_info, "com.sun.star.form.component.DataForm"):
                        print(f'{tab_str}Data From "{name}"')
                    else:
                        print(f'{tab_str}Form "{name}"')
                        # mInfo.Info.show_services("Form", serv_info)
                        # mInfo.Info.show_interfaces("Form", serv_info)
                    child_con = mLo.Lo.qi(XNameAccess, serv_info, True)
                    # recursively list form components
                    cls.list_forms(child_con, tab_str=f"{tab_str}  ")
                elif serv_info.supportsService("com.sun.star.form.FormComponent"):
                    model = mLo.Lo.qi(XControlModel, serv_info)
                    print(f'{tab_str}"{name}":{cls.get_type_str(model)}')
                    #  mProps.Props.show_obj_props("Model", model)
                else:
                    print(f'{tab_str}unknown: "{name}"')
            except Exception:
                print(f'{tab_str}Could not access "{name}"')

    # endregion access forms in document

    # region    get form models
    @overload
    @classmethod
    def get_models(cls, obj: XComponent) -> List[XControlModel]:
        ...

    @overload
    @classmethod
    def get_models(cls, obj: XNameAccess) -> List[XControlModel]:
        ...

    @classmethod
    def get_models(cls, obj: XComponent | XNameAccess) -> List[XControlModel]:
        """
        Gets models from obj

        Args:
            obj (XComponent | XNameAccess): Component or Name Access

        Returns:
            List[XControlModel]: List of found models

        See Also:
            :py:meth:`~.forms.Forms.get_models2`
        """
        if mLo.Lo.is_uno_interfaces(obj, XComponent):
            xcontainer = cls.get_forms(obj)
        else:
            xcontainer = cast(XNameContainer, obj)
        models: List[XControlModel] = []
        nms = xcontainer.getElementNames()
        for name in nms:
            try:
                serv_info = mLo.Lo.qi(XServiceInfo, xcontainer.getByName(name))
                if serv_info.supportsService("com.sun.star.form.FormComponents"):
                    # this means that a form has been found
                    child_con = mLo.Lo.qi(XNameAccess, serv_info, True)
                    # recursively search
                    models.extend(cls.get_models(child_con))
                elif serv_info.supportsService("com.sun.star.form.FormComponent"):
                    model = mLo.Lo.qi(XControlModel, serv_info, True)
                    models.append(model)

            except Exception as e:
                mLo.Lo.print(f'Could not access "{name}"')
        return models

    @classmethod
    def get_models2(cls, doc: XComponent, form_name: str) -> List[XControlModel]:
        """
        Gets models from doc

        Args:
            doc (XComponent): Component
            form_name (str): form name.

        Returns:
            List[XControlModel]: List of found models

        See Also:
            :py:meth:`~.forms.Forms.get_models`
        """
        # another way to obtain models, via the control shapes in the DrawPage
        models: List[XControlModel] = []
        try:
            xdraw_page = cls.get_draw_page(doc)
            if xdraw_page is None:
                return models
        except Exception:
            mLo.Lo.print("No draw page found")
            return models
        try:
            for i in range(xdraw_page.getCount()):
                cshape = mLo.Lo.qi(XControlShape, xdraw_page.getByIndex(i))
                model = cshape.getControl()
                if cls.belongs_to_form(model, form_name):
                    models.append(model)
        except Exception as e:
            mLo.Lo.print(f"Could not collect control model: {e}")
        mLo.Lo.print(f"No. of control models found: {len(models)}")
        return models

    @classmethod
    def get_event_source_name(cls, event: EventObject) -> str:
        """
        Gets event source name

        Args:
            event (EventObject): event object

        Returns:
            str: event source name
        """
        control = mLo.Lo.qi(XControl, event.Source, True)
        return cls.get_name(control)

    @staticmethod
    def get_event_control_model(event: EventObject) -> XControlModel:
        """
        Gets event control model

        Args:
            event (EventObject): event object

        Returns:
            XControlModel: Event control model
        """
        control = mLo.Lo.qi(XControl, event.Source, True)
        return control.getModel()

    @staticmethod
    def get_form_name(ctl_model: XControlModel) -> str:
        """
        Gets form name

        Args:
            ctl_model (XControlModel): control model

        Returns:
            str: form name
        """
        child = mLo.Lo.qi(XChild, ctl_model)
        named = mLo.Lo.qi(XNamed, child.getParent())
        return named.getName()

    @classmethod
    def belongs_to_form(cls, ctl_model: XControlModel, form_name: str) -> bool:
        """
        Get if a control belongs to a form

        Args:
            ctl_model (XControlModel): Control Model
            form_name (str): Form name

        Returns:
            bool: ``True`` if belongs to form; Otherwise, ``False``
        """
        return cls.get_form_name(ctl_model) == form_name

    @staticmethod
    def get_name(ctl_model: XControlModel) -> str:
        """
        Gets name of a given form component

        Args:
            ctl_model (XControlModel): Control Model

        Returns:
            str: Name of component
        """
        return str(mProps.Props.get(ctl_model, "Name"))

    @staticmethod
    def get_label(ctl_model: XControlModel) -> str:
        """
        Gets label of a given form component

        Args:
            ctl_model (XControlModel): Control Model

        Returns:
            str: Label of component
        """
        return str(mProps.Props.get(ctl_model, "Label"))

    @classmethod
    def get_type_str(cls, ctl_model: XControlModel) -> str | None:
        """
        Gets type as string

        Args:
            ctl_model (XControlModel): Control Model

        Returns:
            str | None: Type as string if found; Otherwise, ``None``
        """
        id = cls.get_id(ctl_model)
        if id == -1:
            return None

        serv_info = mLo.Lo.qi(XServiceInfo, ctl_model)
        if id == FormComponentType.COMMANDBUTTON:
            return "Command button"
        elif id == FormComponentType.RADIOBUTTON:
            return "Radio button"
        elif id == FormComponentType.IMAGEBUTTON:
            return "Image button"
        elif id == FormComponentType.CHECKBOX:
            return "Check Box"
        elif id == FormComponentType.LISTBOX:
            return "List Box"
        elif id == FormComponentType.COMBOBOX:
            return "Combo Box"
        elif id == FormComponentType.GROUPBOX:
            return "Group Box"
        elif id == FormComponentType.FIXEDTEXT:
            return "Fixed Text"
        elif id == FormComponentType.GRIDCONTROL:
            return "Grid Control"
        elif id == FormComponentType.FILECONTROL:
            return "File Control"
        elif id == FormComponentType.HIDDENCONTROL:
            return "Hidden Control"
        elif id == FormComponentType.IMAGECONTROL:
            return "Image Control"
        elif id == FormComponentType.DATEFIELD:
            return "Date Field"
        elif id == FormComponentType.TIMEFIELD:
            return "Time Field"
        elif id == FormComponentType.NUMERICFIELD:
            return "Numeric Field"
        elif id == FormComponentType.CURRENCYFIELD:
            return "Currency Field"
        elif id == FormComponentType.PATTERNFIELD:
            return "Pattern Field"
        elif id == FormComponentType.TEXTFIELD:
            # two services with this class id: text field and formatted field
            if serv_info is not None and serv_info.supportsService("com.sun.star.form.component.FormattedField"):
                return "Formatted Field"
            else:
                return "Text Field"
        else:
            mLo.Lo.print(f"Unknown class ID: {id}")
            return None

    @staticmethod
    def get_id(ctl_model: XControlModel) -> int:
        """
        Gets class id for a form component

        Args:
            ctl_model (XControlModel): Control Model

        Returns:
            int: Class Id if found, Otherwise ``-1``
        """
        class_id = mProps.Props.get(ctl_model, "ClassId")
        if class_id is None:
            mLo.Lo.print("No class ID found for form component")
            return -1
        return int(class_id)

    @classmethod
    def is_button(cls, ctl_model: XControlModel) -> bool:
        """
        Gets if component is a command button or a image button

        Args:
            ctl_model (XControlModel): Control Model

        Returns:
            bool: ``True`` if is button; Otherwise, ``False``
        """
        id = cls.get_id(ctl_model)
        if id == -1:
            return False

        return id == FormComponentType.COMMANDBUTTON or id == FormComponentType.IMAGEBUTTON

    @classmethod
    def is_text_field(cls, ctl_model: XControlModel) -> bool:
        """
        Gets if component is a text field

        Args:
            ctl_model (XControlModel): Control Model

        Returns:
            bool: ``True`` if is text field; Otherwise, ``False``
        """
        id = cls.get_id(ctl_model)
        if id == -1:
            return False

        return (
            id == FormComponentType.DATEFIELD
            or id == FormComponentType.TIMEFIELD
            or id == FormComponentType.NUMERICFIELD
            or id == FormComponentType.CURRENCYFIELD
            or id == FormComponentType.PATTERNFIELD
            or id == FormComponentType.TEXTFIELD
        )

    @classmethod
    def is_box(cls, ctl_model: XControlModel) -> bool:
        """
        Gets if component is a box

        Args:
            ctl_model (XControlModel): Control Model

        Returns:
            bool: ``True`` if is box; Otherwise, ``False``
        """
        id = cls.get_id(ctl_model)
        if id == -1:
            return False

        return id == FormComponentType.RADIOBUTTON or id == FormComponentType.CHECKBOX

    @classmethod
    def is_list(cls, ctl_model: XControlModel) -> bool:
        """
        Gets if component is a list

        Args:
            ctl_model (XControlModel): Control Model

        Returns:
            bool: ``True`` if is list; Otherwise, ``False``
        """
        id = cls.get_id(ctl_model)
        if id == -1:
            return False

        return id == FormComponentType.LISTBOX or id == FormComponentType.COMBOBOX

    # Other control types
    # FormComponentType.GROUPBOX
    # FormComponentType.FIXEDTEXT
    # FormComponentType.GRIDCONTROL
    # FormComponentType.FILECONTROL
    # FormComponentType.HIDDENCONTROL
    # FormComponentType.IMAGECONTROL
    # FormComponentType.SCROLLBAR
    # FormComponentType.SPINBUTTON
    # FormComponentType.NAVIGATIONBAR

    # endregion get form models

    # region    get control for a model
    @staticmethod
    def get_control(doc: XComponent, ctl_model: XControlModel) -> XControl:
        """
        Gets the control from the specified control model.

        Args:
            doc (XComponent): Component
            ctl_model (XControlModel): Control Model

        Raises:
            Exception: If unable to get control

        Returns:
            XControl: Control
        """
        try:
            control_access = mGui.GUI.get_control_access(doc)
            if control_access is None:
                raise Exception("Could not obtain controls access in document")
            return control_access.getControl(ctl_model)
        except Exception as e:
            raise Exception(f"Could not access control: {e}") from e

    @classmethod
    def get_named_control(cls, doc: XComponent, ctl_name: str) -> XControl | None:
        """
        Gets a named control.

        Args:
            doc (XComponent): Component
            ctl_name (str): Name of control

        Returns:
            XControl | None: Control if found; Otherwise, None
        """
        models = cls.get_models(doc)
        ctl = None
        for model in models:
            if cls.get_name(model) == ctl_name:
                mLo.Lo.print(f"Found: {ctl_name}")
                try:
                    ctl = cls.get_control(doc, model)
                except Exception as e:
                    mLo.Lo.print("Error getting control.")
                    mLo.Lo.print(f"  {e}")
                    ctl = None
                finally:
                    break
        return ctl

    @classmethod
    def get_control_model(cls, doc: XComponent, ctl_name) -> XControlModel | None:
        """
        Gets Control Model by Name

        Args:
            doc (XComponent): Component
            ctl_name (str): Name of control

        Returns:
            XControlModel | None: Control Model if found; Otherwise, None
        """
        control = cls.get_named_control(doc, ctl_name)
        if control is None:
            return None
        return control.getModel()

    # endregion get control for a model

    # region create controls

    # region    add_control
    @overload
    @classmethod
    def add_control(
        cls,
        doc: XComponent,
        name: str,
        label: str | None,
        comp_kind: FormComponentKind | str,
        x: int,
        y: int,
        width: int,
        height: int,
    ) -> XPropertySet:
        ...

    @overload
    @classmethod
    def add_control(
        cls,
        doc: XComponent,
        name: str,
        label: str | None,
        comp_kind: FormComponentKind | str,
        x: int,
        y: int,
        width: int,
        height: int,
        anchor_type: TextContentAnchorType | None = None,
        parent_form: XNameContainer | None = None,
    ) -> XPropertySet:
        ...

    @classmethod
    def add_control(
        cls,
        doc: XComponent,
        name: str,
        label: str | None,
        comp_kind: FormComponentKind | str,
        x: int,
        y: int,
        width: int,
        height: int,
        anchor_type: TextContentAnchorType | None = None,
        parent_form: XNameContainer | None = None,
    ) -> XPropertySet:
        """
        Add a control

        Args:
            doc (XComponent): Component
            name (str): Control Name
            label (str | None): Label to assign to control
            comp_kind (FormComponentKind | str): Kind of control such as ``CheckBox``.
            x (int): Control X position
            y (int): Control Y Position
            width (int): Control width#
            height (int): control height
            anchor_type (TextContentAnchorType | None): Control Anchor Type. Defaults to ``TextContentAnchorType.AT_PARAGRAPH``
            parent_form (XNameContainer | None): Parent form in which to add control.

        Returns:
            XPropertySet: Control Property Set

        See Also:
            For ``comp_kind`` `API component Module Namespace <https://api.libreoffice.org/docs/idl/ref/namespacecom_1_1sun_1_1star_1_1form_1_1component.html>`_
        """
        try:
            # create a shape to represent the control's view
            #     xDocAsFactory = mLo.Lo.qi(XMultiServiceFactory, doc, True)
            #     cshpae =
            #     XMultiServiceFactory xDocAsFactory = (XMultiServiceFactory)UnoRuntime.queryInterface(
            #   XMultiServiceFactory.class, s_aDocument);

            #     XControlShape xShape = (XControlShape)UnoRuntime.queryInterface(XControlShape.class,
            #   xDocAsFactory.createInstance("com.sun.star.drawing.ControlShape"));

            cshape = mLo.Lo.create_instance_msf(XControlShape, "com.sun.star.drawing.ControlShape", raise_err=True)

            # position and size of the shape
            cshape.setSize(Size(width * 100, height * 100))
            cshape.setPosition(Point(x * 100, y * 100))

            # adjust the anchor so that the control is tied to the page
            shape_props = mLo.Lo.qi(XPropertySet, cshape, True)

            if anchor_type is None:
                shape_props.setPropertyValue("AnchorType", TextContentAnchorType.AT_PARAGRAPH)
            else:
                shape_props.setPropertyValue("AnchorType", TextContentAnchorType(anchor_type))

            # create the control's model, this is a service
            # see: https://api.libreoffice.org/docs/idl/ref/servicecom_1_1sun_1_1star_1_1form_1_1FormControlModel.html
            cmodel = mLo.Lo.create_instance_mcf(XControlModel, f"com.sun.star.form.component.{comp_kind}")

            # insert the model into the form (or default to "Form")
            if parent_form is not None:
                parent_form.insertByName(name, cmodel)

            # link model to the shape
            cshape.setControl(cmodel)

            # add the shape to the shapes on the doc's draw page
            draw_page = cls.get_draw_page(doc)
            form_shapes = mLo.Lo.qi(XShapes, draw_page)
            form_shapes.add(cshape)

            # set Name and Label properties for the model
            model_props = mLo.Lo.qi(XPropertySet, cmodel, True)
            model_props.setPropertyValue("Name", name)
            if label is not None:
                model_props.setPropertyValue("Label", label)
            return model_props
        except Exception:
            raise
        # endregion add_control

    # region    add_labelled_control

    @overload
    @classmethod
    def add_labelled_control(
        cls, doc: XComponent, label: str, comp_kind: FormComponentKind | str, y: int
    ) -> XPropertySet:
        ...

    @overload
    @classmethod
    def add_labelled_control(
        cls, doc: XComponent, label: str, comp_kind: FormComponentKind | str, x: int, y: int, height: int
    ) -> XPropertySet:
        ...

    @classmethod
    def add_labelled_control(cls, *args, **kwargs) -> None:
        """
        Create a label and data field control, with the label preceding the control

        Args:
            doc (XComponent): Component
            label (str): Label to assign to control
            comp_kind (FormComponentKind | str): Kind of control such as ``CheckBox``.
            x (int): Control X position
            y (int): Control Y Position
            height (int): control height

        Returns:
            XPropertySet: DataField Control Property Set
        """
        ordered_keys = (1, 2, 3, 4, 5, 6)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("doc", "label", "comp_kind", "x", "y", "height")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("addLabelledControl() got an unexpected keyword argument")
            ka[1] = kwargs.get("doc", None)
            ka[2] = kwargs.get("label", None)
            ka[3] = kwargs.get("comp_kind", None)
            keys = ("x", "y")
            for key in keys:
                if key in kwargs:
                    ka[4] = kwargs[key]
                    break
            if count == 4:
                return ka
            ka[5] = kwargs.get("y", None)
            ka[6] = kwargs.get("height", None)
            return ka

        if not count in (4, 6):
            raise TypeError("addLabelledControl() got an invalid numer of arguments")

        kargs = get_kwargs()

        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        if count == 4:
            return cls._add_labelled_control(
                doc=kargs[1], label=kargs[2], comp_kind=kargs[3], x=2, y=kargs[4], height=6
            )
        else:
            return cls._add_labelled_control(
                doc=kargs[1], label=kargs[2], comp_kind=kargs[3], x=kargs[4], y=kargs[5], height=kargs[6]
            )

    @classmethod
    def _add_labelled_control(
        cls, doc: XComponent, label: str, comp_kind: FormComponentKind | str, x: int, y: int, height: int
    ) -> XPropertySet:
        try:
            name = f"{label}_label"
            # create label (fixed text) control
            label_props = cls.add_control(
                doc=doc,
                name=name,
                label=label,
                comp_kind=FormComponentKind.FIXED_TEXT,
                x=x,
                y=y,
                width=40,
                height=height,
            )

            # create data field control
            ctl_props = cls.add_control(
                doc=doc,
                name=label,
                label=None,
                comp_kind=comp_kind,
                x=x + 26,
                y=y,
                width=40,
                height=height,
            )
            ctl_props.setPropertyValue("DataField", label)

            # add label props to the control
            ctl_props.setPropertyValue("LabelControl", label_props)
            return ctl_props
        except Exception:
            raise

    # endregion add_labelled_control

    # region    add_button
    @overload
    @classmethod
    def add_button(cls, doc: XComponent, name: str, label: str | None, x: int, y: int, width: int) -> XPropertySet:
        ...

    @overload
    @classmethod
    def add_button(
        cls, doc: XComponent, name: str, label: str | None, x: int, y: int, width: int, height: int
    ) -> XPropertySet:
        ...

    @classmethod
    def add_button(
        cls, doc: XComponent, name: str, label: str | None, x: int, y: int, width: int, height: int = 6
    ) -> XPropertySet:
        """
        Adds a button control.

        By Default the button has no tab stop and does not focus on click.

        Args:
            doc (XComponent): Component
            name (str): Button name
            label (str | None): Button Label
            x (int): Button X position
            y (int): Button Y position
            height (int): Button Height
            width (int, optional): Button Height. Defaults to 6.

        Returns:
            XPropertySet: Button Property Set
        """
        try:
            btn_props = cls.add_control(
                doc=doc,
                name=name,
                label=label,
                comp_kind=FormComponentKind.COMMAND_BUTTON,
                x=x,
                y=y,
                width=width,
                height=height,
            )
            # don't want button to be accessible by the "tab" key
            btn_props.setPropertyValue("Tabstop", False)

            # the button should not steal focus when clicked
            btn_props.setPropertyValue("FocusOnClick", False)

            return btn_props
        except Exception:
            raise

    # endregion add_button

    @classmethod
    def add_list(
        cls, doc: XComponent, name: str, entries: Iterable[str], x: int, y: int, width: int, height: int
    ) -> XPropertySet:
        """
        Adds a list

        Args:
            doc (XComponent): Component
            name (str): List Name
            entries (Iterable[str]): List Entries
            x (int): List X position
            y (int): List Y Position
            width (int): List Width
            height (int): List Height

        Returns:
            XPropertySet: List property set
        """
        try:
            lst_props = cls.add_control(
                doc=doc,
                name=name,
                label=None,
                comp_kind=FormComponentKind.LIST_BOX,
                x=x,
                y=y,
                width=width,
                height=height,
            )
            items = mProps.Props.any(*[s for s in entries])
            # lst_props.setPropertyValue("DefaultSelection", 0)
            uno.invoke(lst_props, "setPropertyValue", ("ListSource", items))
            uno.invoke(lst_props, "setPropertyValue", ("DefaultSelection", mProps.Props.any(0)))
            lst_props.setPropertyValue("Dropdown", True)
            lst_props.setPropertyValue("MultiSelection", False)
            uno.invoke(lst_props, "setPropertyValue", ("StringItemList", items))
            uno.invoke(lst_props, "setPropertyValue", ("SelectedItems", mProps.Props.any(0)))
            return lst_props
        except Exception:
            raise

    @classmethod
    def add_database_list(
        cls, doc: XComponent, name: str, sql_cmd: str, x: int, y: int, width: int, height: int
    ) -> XPropertySet:
        """
        Add a list with a SQL command as it data source

        Args:
            doc (XComponent): Component
            name (str): List Name
            sql_cmd (str): SQL Command
            x (int): List X position
            y (int): List Y Position
            width (int): List Width
            height (int): List Height

        Returns:
            XPropertySet: List property set
        """
        try:
            lst_props = cls.add_control(
                doc=doc,
                name=name,
                label=None,
                comp_kind=FormComponentKind.DATABASE_LIST_BOX,
                x=x,
                y=y,
                width=width,
                height=height,
            )
            lst_props.setPropertyValue("Dropdown", True)
            lst_props.setPropertyValue("MultiSelection", False)
            lst_props.setPropertyValue("BoundColumn", 0)

            # data-aware properties
            lst_props.setPropertyValue("ListSourceType", ListSourceType.SQL)
            uno.invoke(lst_props, "setPropertyValue", ("ListSource", mProps.Props.any(sql_cmd)))
            return lst_props
        except Exception:
            raise

    @staticmethod
    def create_grid_column(grid_model: XControlModel, data_field: str, col_kind: str, width: int) -> None:
        """
        Adds a column to a gird

        Args:
            grid_model (XControlModel): Grid control Model
            data_field (str): the database field to which the column should be bound
            col_kind (str):  the column type such as "NumericField"
            width (int): the column width (in mm). If 0, no width is set.

        Returns:
            None:
        """
        # column container and factory
        col_container = mLo.Lo.qi(XIndexContainer, grid_model)
        col_factory = mLo.Lo.qi(XGridColumnFactory, grid_model)

        # create the column
        col_props = col_factory.createColumn(col_kind)
        col_props.setPropertyValue("DataField", data_field)
        col_props.setPropertyValue("Label", data_field)
        col_props.setPropertyValue("Name", data_field)
        if width > 0:
            col_props.setPropertyValue("Width", width * 10)

        # add properties column to container
        col_container.insertByIndex(col_container.getCount(), col_props)

    # endregion create controls

    # region  bind form to database
    @staticmethod
    def bind_form_to_table(xform: XForm, src_name: str, tbl_name: str) -> None:
        """
        Bind the form to the database in the src_name URL

        Args:
            xform (XForm): Form
            src_name (str): Source Name URL
            tbl_name (str): Table Name

        Returns:
            None:
        """
        mProps.Props.set(xform, DataSourceName=src_name, Command=tbl_name, CommandType=CommandType.TABLE)

    @staticmethod
    def bind_form_to_sql(xform: XForm, src_name: str, cmd: str) -> None:
        """
        Bind the form to the database in the ``src_name`` URL, and send a SQL cmd

        Args:
            xform (XForm): Form
            src_name (str): Source Name URL
            cmd (str): Command

        Returns:
            None:
        """
        mProps.Props.set(xform, DataSourceName=src_name, Command=cmd, CommandType=CommandType.COMMAND)
        # cannot use CommandType.TABLE for the SELECT cmd

    # endregion  bind form to database

    # region  bind a macro to a form control
    @staticmethod
    def assign_script(
        ctl_props: XPropertySet, interface_name: str, method_name: str, script_name: str, loc: str
    ) -> None:
        """
        Binds a macro to a form control

        Args:
            ctl_props (XPropertySet): Properties of control
            interface_name (str): Interface Name
            method_name (str): Method Name
            script_name (str): Script Name
            loc (str): can be user, share, document, and extensions

        Returns:
            None:

        See Also:
            `Scripting Framework URI Specification <https://wiki.openoffice.org/wiki/Documentation/DevGuide/Scripting/Scripting_Framework_URI_Specification>`_
        """
        # https://wiki.openoffice.org/wiki/Documentation/DevGuide/WritingUNO/XInterface
        # In C++, two objects are the same if their XInterface are the same. The queryInterface() for XInterface will have to
        # be called on both. In Java, check for the identity by calling the runtime function
        # com.sun.star.uni.UnoRuntime.areSame().
        try:
            props_child = mLo.Lo.qi(XChild, ctl_props, True)
            parent_form = mLo.Lo.qi(XIndexContainer, props_child.getParent(), True)

            pos = -1
            for i in range(parent_form.getCount()):
                child = mLo.Lo.qi(XPropertySet, parent_form.getByIndex(i))
                if mInfo.Info.is_same(child, ctl_props):
                    pos = i
                    break

            if pos == -1:
                mLo.Lo.print("Could not find contol's position in form")
            else:
                mgr = mLo.Lo.qi(XEventAttacherManager, parent_form, True)
                ed = ScriptEventDescriptor(
                    interface_name,
                    method_name,
                    "",
                    "Script",
                    f"vnd.sun.star.script:{script_name}?language=Java&location={loc}",
                )

                mgr.registerScriptEvent(pos, ed)
        except Exception:
            raise

    # endregion  bind a macro to a form control



from enum import Enum
from . import kind_helper


class DrawingGradientKind(str, Enum):
    """
    Drawing (some) Gradient Values.
    The actual value may change from version to version and may be different
    in other languages.

    These are just quick suggestion that show up in Draw Gradient colors.
    """

    BLANK_WITH_GRAY = "Blank with Gray"
    BLUE_TOUCH = "Blue Touch"
    DEEP_OCEAN = "Deep Ocean"
    GREEN_GRASS = "Green Grass"
    LONDON_MIST = "London Mist"
    MAHOGANY = "Mahogany"
    MIDNIGHT = "Midnight"
    NEON_LIGHT = "Neon Light"
    PASTEL_BOUQUET = "Pastel Bouquet"
    PASTEL_DREAM = "Pastel Dream"
    PRESENT = "Present"
    SPOTTED_GRAY = "Spotted Gray"
    SUBMARINE = "Submarine"
    SUNSHINE = "Sunshine"
    TEAL_TO_BLUE = "Teal to Blue"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_str(s: str) -> "DrawingGradientKind":
        """
        Gets an ``DrawingGradientKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``DrawingGradientKind`` instance.

        Returns:
            DrawingGradientKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, DrawingGradientKind)



from enum import IntEnum
from . import kind_helper


class SearchByKind(IntEnum):
    """
    Gallery Search by Kind.

    Used to determine search criteria.

    See Also:
        :py:meth:`.Gallery.find_gallery_item`
    """

    FILE_NAME = 1
    """Match File Name"""
    TITLE = 2
    """Match Title"""

    @staticmethod
    def from_str(s: str) -> "SearchByKind":
        """
        Gets an ``SearchByKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``SearchByKind`` instance.

        Returns:
            SearchByKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, SearchByKind)



"""
This module is for mocking unohelper.py

DO NOT! Use this class unless you are certain what you are doing!

Read the Docs gives issues with unohelper.Base for classes that inherit from it.

The solution is to Mock unohelper during read the docs build.
"""
# it is necessary to fence using DOCS_BUILDING here.
# if not stickytape will import sphinx modules.

from .mock_g import DOCS_BUILDING
if DOCS_BUILDING:
    from sphinx.ext.autodoc.mock import _MockObject
    class Base(_MockObject):
        pass
else:
    import unohelper
    class Base(unohelper.Base):
        pass


from __future__ import annotations
from abc import abstractclassmethod
from enum import Enum


class ChartBaseTypeEnum(str, Enum):
    def __str__(self) -> str:
        return str(self.value)

    @abstractclassmethod
    def to_namespace(self) -> str:
        ...


class ChartTemplateBase(ChartBaseTypeEnum):
    """
    Base Enum for all enums found in :py:class:`~.chart2_types.ChartTypes`

    Example:

        :py:attr:`.Chart2.ChartLookup` is an alias of :py:class:`~.chart2_types.ChartTypes`

        .. code-block:: python

            Chart2.has_categories(diagram_name=Chart2.ChartLookup.Bar.TYPE_PERCENT.BAR_DEEP_3D)

    See Also:
        :py:meth:`.Chart2.has_categories`
    """

    def to_namespace(self) -> str:
        return f"com.sun.star.chart2.template.{self.value}"


class ChartTypeNameBase(ChartBaseTypeEnum):
    def to_namespace(self) -> str:
        return f"com.sun.star.chart2.{self.value}"


class ColumnStackedKind(ChartTemplateBase):
    COLUMN = "Column"
    STACKED_COLUMN = "StackedColumn"
    PERCENT_STACKED_COLUMN = "PercentStackedColumn"
    COLUMN_WITH_LINE = "ColumnWithLine"
    STACKED_COLUMN_WITH_LINE = "StackedColumnWithLine"


class ColumnPercentKind(ChartTemplateBase):
    COLUMN_DEEP_3D = "ThreeDColumnDeep"
    COLUMN_FLAT_3D = "ThreeDColumnFlat"


class Column3dKind(ChartTemplateBase):
    STACKED_3D_COLUMN_FLAT = "StackedThreeDColumnFlat"
    PERCENT_STACKED_3D_COLUMN_FLAT = "PercentStackedThreeDColumnFlat"


class BarStackedKind(ChartTemplateBase):
    BAR = "Bar"
    STACKED_BAR = "StackedBar"
    PERCENT_STACKED_BAR = "PercentStackedBar"


class BarPercentKind(ChartTemplateBase):
    BAR_DEEP_3D = "ThreeDBarDeep"
    BAR_FLAT_3D = "ThreeDBarFlat"


class Bar3dKind(ChartTemplateBase):
    STACKED_3D_BAR_FLAT = "StackedThreeDBarFlat"
    PERCENT_STACKED_3D_BAR_FLAT = "PercentStackedThreeDBarFlat"


class PieDonutKind(ChartTemplateBase):
    PIE = "Pie"
    DONUT = "Donut"


class PieExplodeKind(ChartTemplateBase):
    DONUT_3D = "ThreeDDonut"
    DONUT_ALL_EXPLODED_3D = "ThreeDDonutAllExploded"


class Pie3dKind(ChartTemplateBase):
    PIE_3D = "ThreeDPie"
    PIE_ALL_EXPLODED_3D = "ThreeDPieAllExploded"
    DONUT_3D = "ThreeDDonut"
    DONUT_ALL_EXPLODED_3D = "ThreeDDonutAllExploded"


class AreaStackedKind(ChartTemplateBase):
    AREA = "Area"
    STACKED_AREA = "StackedArea"
    PERCENT_STACKED_AREA = "PercentStackedArea"


class AreaPercentKind(ChartTemplateBase):
    AREA_3D = "ThreeDArea"
    STACKED_AREA_3D = "StackedThreeDArea"


class Area3dKind(ChartTemplateBase):
    PERCENT_STACKED_AREA_3D = "PercentStackedThreeDArea"


class LineSymbolKind(ChartTemplateBase):
    LINE = "Line"
    SYMBOL = "Symbol"
    LINE_SYMBOL = "LineSymbol"


class LineStackedKind(ChartTemplateBase):
    STACKED_LINE = "StackedLine"
    STACKED_SYMBOL = "StackedSymbol"
    STACKED_LINE_SYMBOL = "StackedLineSymbol"


class LinePercentKind(ChartTemplateBase):
    PERCENT_STACKED_LINE = "PercentStackedLine"
    PERCENT_STACKED_SYMBOL = "PercentStackedSymbol"


class Line3dKind(ChartTemplateBase):
    PERCENT_STACKED_LINE_SYMBOL = "PercentStackedLineSymbol"
    LINE_3D = "ThreeDLine"
    LINE_DEEP_3D = "ThreeDLineDeep"
    STACKED_LINE_3D = "StackedThreeDLine"
    PERCENT_STACKED_LINE_3D = "PercentStackedThreeDLine"


class XYLineKind(ChartTemplateBase):
    SCATTER_SYMBOL = "ScatterSymbol"
    SCATTER_LINE = "ScatterLine"
    SCATTER_LINE_SYMBOL = "ScatterLineSymbol"


class XY3dKind(ChartTemplateBase):
    SCATTER_3D = "ThreeDScatter"


class BubbleKind(ChartTemplateBase):
    BUBBLE = "Bubble"


class NetLineKind(ChartTemplateBase):
    NET = "Net"
    NET_LINE = "NetLine"
    NET_SYMBOL = "NetSymbol"
    FILLED_NET = "FilledNet"


class NetSymbolKind(ChartTemplateBase):
    STACKED_NET = "StackedNet"
    STACKED_NET_LINE = "StackedNetLine"


class NetFilledKind(ChartTemplateBase):
    STACKED_NET_SYMBOL = "StackedNetSymbol"
    STACKED_FILLED_NET = "StackedFilledNet"


class NetStackedKind(ChartTemplateBase):
    PERCENT_STACKED_NET = "PercentStackedNet"
    PERCENT_STACKED_NET_LINE = "PercentStackedNetLine"
    PERCENT_STACKED_NET_SYMBOL = "PercentStackedNetSymbol"


class NetPercentKind(ChartTemplateBase):
    PERCENT_STACKED_FILLED_NET = "PercentStackedFilledNet"


class StockOpenKind(ChartTemplateBase):
    STOCK_LOW_HIGH_CLOSE = "StockLowHighClose"


class StockVolumeKind(ChartTemplateBase):
    STOCK_OPEN_LOW_HIGH_CLOSE = "StockOpenLowHighClose"
    STOCK_VOLUME_LOW_HIGH_CLOSE = "StockVolumeLowHighClose"
    STOCK_VOLUME_OPEN_LOW_HIGH_CLOSE = "StockVolumeOpenLowHighClose"


class NamedColumnKind(ChartTypeNameBase):
    COLUMN_CHART = "ColumnChartType"


class NamedBarKind(ChartTypeNameBase):
    BAR_CHART = "BarChartType"


class NamedPieKind(ChartTypeNameBase):
    PIE_CHART = "PieChartType"


class NamedAreaKind(ChartTypeNameBase):
    AREA_CHART = "AreaChartType"


class NamedLineKind(ChartTypeNameBase):
    LINE_CHART = "LineChartType"


class NamedXYKind(ChartTypeNameBase):
    SCATTER_CHART = "ScatterChartType"


class NamedBubbleKind(ChartTypeNameBase):
    BUBBLE_CHART = "BubbleChartType"


class NamedNetKind(ChartTypeNameBase):
    NET_CHART = "NetChartType"
    FILLED_NET_CHART = "FilledNetChartType"


class NamedStockKind(ChartTypeNameBase):
    CANDLE_STICK_CHART = "CandleStickChartType"


class ChartTypes:
    """
    Class for convenient lookup of chart type names.
    """

    class Column:
        NAMED = NamedColumnKind
        TEMPLATE_3D = Column3dKind
        TEMPLATE_PERCENT = ColumnPercentKind
        TEMPLATE_STACKED = ColumnStackedKind

    class Bar:
        NAMED = NamedBarKind
        TEMPLATE_3D = Bar3dKind
        TEMPLATE_PERCENT = BarPercentKind
        TEMPLATE_STACKED = BarStackedKind

    class Pie:
        NAMED = NamedPieKind
        TEMPLATE_3D = Pie3dKind
        TEMPLATE_DONUT = PieDonutKind
        TEMPLATE_EXPLODE = PieExplodeKind

    class Area:
        NAMED = NamedAreaKind
        TEMPLATE_3D = Area3dKind
        TEMPLATE_PERCENT = AreaPercentKind
        TEMPLATE_STACKED = AreaStackedKind

    class Line:
        NAMED = NamedLineKind
        TEMPLATE_3D = Line3dKind
        TEMPLATE_PERCENT = LinePercentKind
        TEMPLATE_STACKED = LineStackedKind
        TEMPLATE_SYMBOL = LineSymbolKind

    class XY:
        NAMED = NamedXYKind
        TEMPLATE_3D = XY3dKind
        TEMPLATE_LINE = XYLineKind

    class Bubble:
        NAMED = NamedBubbleKind
        TEMPLATE_BUBBLE = BubbleKind

    class Net:
        NAMED = NamedNetKind
        TEMPLATE_FILLED = NetFilledKind
        TEMPLATE_LINE = NetLineKind
        TEMPLATE_PERCENT = NetPercentKind
        TEMPLATE_STACKED = NetStackedKind
        TEMPLATE_SYMBOL = NetSymbolKind

    class Stock:
        NAMED = NamedStockKind
        TEMPLATE_OPEN = StockOpenKind
        TEMPLATE_VOLUME = StockVolumeKind



# coding: utf-8
# Origin class https://gitlab.com/LibreOfficiant/ide_utils/-/blob/master/IDE_utils.py
# Original Author: LibreOfficiant: https://gitlab.com/LibreOfficiant
from __future__ import annotations
from typing import TYPE_CHECKING
import uno
import os
_ON_RTD = os.environ.get('READTHEDOCS', None) == 'True'
if _ON_RTD:
    from ..mock import unohelper
else:
    import unohelper

from com.sun.star.script.provider import XScriptContext

if TYPE_CHECKING:
    from com.sun.star.frame import XDesktop
    from com.sun.star.frame import XModel
    from com.sun.star.uno import XComponentContext


class ScriptContext(unohelper.Base, XScriptContext):
    """Substitute (Libre|Open)Office XSCRIPTCONTEXT built-in

    Can be used in IDEs such as Anaconda, Geany, KDevelop, PyCharm, VS Code..
    in order to run/debug Python macros.

    Implements: com.sun.star.script.provider.XScriptContext
    """

    def __init__(self, ctx, desktop, doc):
        self.ctx = ctx
        self.desktop = desktop
        self.doc = doc

    def getComponentContext(self) -> XComponentContext:
        return self.ctx

    def getDesktop(self) -> XDesktop:
        # return self.ctx.getServiceManager().createInstanceWithContext("com.sun.star.frame.Desktop", self.ctx)
        if self.desktop is None:
            try:
                self.desktop = self.ctx.getServiceManager().createInstanceWithContext("com.sun.star.frame.Desktop", self.ctx)
            except Exception:
                self.desktop = None
        return self.desktop

    def getDocument(self) -> XModel:
        # return self.getDesktop().getCurrentComponent()
        if self.doc is None:
            try:
                self.doc =self.getDesktop().getCurrentComponent()
            except Exception:
                self.doc = None
        return self.doc

    def getInvocationContext(self):
        raise os.NotImplementedError



# coding: utf-8
# Python conversion of ViewState.java by Andrew Davison, ad@fivedots.coe.psu.ac.th
# See Also: https:#fivedots.coe.psu.ac.th/~ad/jlop/
from __future__ import annotations
from enum import IntEnum
from ..office import calc as mCalc


class ViewState:
    """for moving the pane focus"""
    class PaneEnum(IntEnum):
        MOVE_UP = 0
        MOVE_DOWN = 1
        MOVE_LEFT = 2
        MOVE_RIGHT = 3

    def __init__(self, state: str) -> None:
        """
        Constructor

        Args:
            state (str): State in format of '0/4998/0/1/0/218/2/0/0/4988/4998'

        Raises:
            ValueError: if state does not contains 10 '/' (11 parts)
        """
        # The state string has the format:
        #       0/4998/0/1/0/218/2/0/0/4988/4998

        self._cursor_column: int = 0
        self._cursor_row: int = 0
        self._col_split_mode: int = 0
        self._col_split_mode: int = 0
        self._row_split_mode: int = 0
        self._row_split_mode: int = 0
        self._vertical_split: int = 0
        self._horizontal_split: int = 0
        self._pane_focus_num: int = 0
        self._column_left_pane: int = 0
        self._column_right_pane: int = 0
        self._row_upper_pane: int = 0
        self._row_lower_pane: int = 0
        states = state.split("/")
        if len(states) != 11:
            raise ValueError(f"Incorrect number of states, Expected 11 got {len(states)}")

        self._cursor_column = ViewState.parse_int(states[0])  # 0: cursor position column
        self._cursor_row = ViewState.parse_int(states[1])  # 1: cursor position row

        self._col_split_mode = ViewState.parse_int(states[2])  # 2: column split mode
        self._row_split_mode = ViewState.parse_int(states[3])  # 3: row split mode

        self._vertical_split = ViewState.parse_int(states[4])  # 4: vertical split position
        self._horizontal_split = ViewState.parse_int(states[5])  # 5: horizontal split position

        self._pane_focus_num = ViewState.parse_int(states[6])  # 6: focused pane number

        self._column_left_pane = ViewState.parse_int(states[7])  # 7: left column index of left pane
        self._column_right_pane = ViewState.parse_int(states[8])  # 8: left column index of right pane

        self._row_upper_pane = ViewState.parse_int(states[9])  # 9: top row index of upper pane
        self._row_lower_pane = ViewState.parse_int(states[10])  # 10: top row index of lower pane

    @staticmethod
    def parse_int(s: str) -> int:
        """
        Parses int

        Args:
            s (str): string value that contains int

        Returns:
            int: string value converted to int on success; Otherwise, 0
        """
        if not s:
            return 0
        try:
            return int(s)
        except ValueError:
            print(f"'{s}' could not be parsed as an int; using 0")
        return 0

    @property
    def cursor_column(self) -> int:
        """Gets/Sets cursor position column"""
        return self._cursor_column

    @cursor_column.setter
    def cursor_column(self, value: int) -> None:
        if value < 0:
            raise ValueError("Column position must be positive")
        self._cursor_column = value

    @property
    def cursor_row(self) -> int:
        """Gets/Sets cursor position row"""
        return self._cursor_row

    @cursor_row.setter
    def cursor_row(self, value: int) -> None:
        if value < 0:
            raise ValueError("Row position must be positive")
        self._cursor_row = value

    @property
    def column_split_mode(self) -> int:
        """Gets/Sets column split mode"""
        return self._col_split_mode

    @column_split_mode.setter
    def column_split_mode(self, value: bool | int) -> None:
        self._col_split_mode = int(value)
        if self._col_split_mode == 0:  # no column splitting
            self._vertical_split = 0
            if self._pane_focus_num == 1 or self._pane_focus_num == 3:
                self._pane_focus_num -= 1  # move focus to left

    @property
    def row_split_mode(self) -> int:
        """Gets/Sets row split mode"""
        return self._row_split_mode

    @row_split_mode.setter
    def row_split_mode(self, value: bool | int) -> None:
        self._row_split_mode = int(value)
        if self._row_split_mode == 0:  # no row splitting
            self._horizontal_split = 0
            if self._pane_focus_num == 2 or self._pane_focus_num == 3:
                self._pane_focus_num -= 2  # move focus up

    @property
    def vertical_split(self) -> int:
        """Gets/Sets vertical split position"""
        return self._vertical_split

    @vertical_split.setter
    def vertical_split(self, value: int) -> None:
        if value < 0:
            raise ValueError("Position must be positive")
        self._vertical_split = value

    @property
    def horizontal_split(self) -> int:
        """Gets/Sets horizontal split position"""
        return self._horizontal_split

    @horizontal_split.setter
    def horizontal_split(self, value: int) -> None:
        if value < 0:
            raise ValueError("Position must be positive")
        self._horizontal_split = value

    @property
    def pane_focus_num(self) -> int:
        """Gets/Sets focused pane number"""
        return self._pane_focus_num

    @pane_focus_num.setter
    def pane_focus_num(self, value: int) -> None:
        if value < 0 or value > 3:
            raise ValueError("Focus number is out of range 0-3")

        if self._horizontal_split == 0 and (value == 1 or value == 3):
            raise ValueError("No horizontal split, so focus number must be 0 or 2")

        if self._vertical_split == 0 and (value == 2 or value == 3):
            raise ValueError("No vertical split, so focus number must be 0 or 1")
        self._pane_focus_num = value

    def move_pane_focus(self, dir: int | ViewState.PaneEnum) -> bool:
        """
        Moves pane focus

        Args:
            dir (int | PaneEnum): Direction to move

        Raises:
            ValueError: If dir is unknown

        Returns:
            bool: True if move is successful; Otherwise False

        Note:
            The 4 possible view panes are numbered like so
            ::

                0  |  1
                -------
                2  |  3

            If there's no horizontal split then the panes are numbered 0 and 2.
            If there's no vertical split then the panes are numbered 0 and 1.
        """
        try:
            d = ViewState.PaneEnum(dir)
        except Exception:
            raise ValueError("Unknown move direction")

        if d == ViewState.PaneEnum.MOVE_UP:
            if self._pane_focus_num == 3:
                self._pane_focus_num = 1
            elif self._pane_focus_num == 2:
                self._pane_focus_num = 0
            else:
                print("cannot move up")
                return False
        elif d == ViewState.PaneEnum.MOVE_DOWN:
            if self._pane_focus_num == 1:
                self._pane_focus_num = 3
            elif self._pane_focus_num == 0:
                self._pane_focus_num = 2
            else:
                print("cannot move down")
                return False
        elif d == ViewState.PaneEnum.MOVE_LEFT:
            if self._pane_focus_num == 1:
                self._pane_focus_num = 0
            elif self._pane_focus_num == 3:
                self._pane_focus_num = 2
            else:
                print("cannot move left")
                return False
        elif d == ViewState.PaneEnum.MOVE_RIGHT:
            if self._pane_focus_num == 0:
                self._pane_focus_num = 1
            elif self._pane_focus_num == 2:
                self._pane_focus_num = 3
            else:
                print("cannot move right")
                return False
        return True

    @property
    def column_left_pane(self) -> int:
        """Gets/Sets left column index of left pane"""
        return self._column_left_pane

    @column_left_pane.setter
    def column_left_pane(self, value: int) -> None:
        if value < 0:
            raise IndexError("value must be positive")
        self._column_left_pane = value

    @property
    def column_right_pane(self) -> int:
        """Gets/Sets left column index of right pane"""
        return self._column_right_pane

    @column_right_pane.setter
    def column_right_pane(self, value: int) -> None:
        if value < 0:
            raise IndexError("value must be positive")
        self._column_right_pane = value

    @property
    def row_upper_pane(self) -> int:
        """Gets/Sets top row index of upper pane"""
        return self._row_upper_pane

    @row_upper_pane.setter
    def row_upper_pane(self, value: int) -> None:
        if value < 0:
            raise IndexError("value must be positive")
        self._row_upper_pane = value

    @property
    def row_lower_pane(self) -> int:
        """Gets/Sets top row index of lower pane"""
        return self._row_lower_pane

    @row_lower_pane.setter
    def row_lower_pane(self, value: int) -> None:
        if value < 0:
            raise IndexError("value must be positive")
        self._row_lower_pane = value

    def report(self) -> None:
        """
        Prints a report to console
        """
        print("Sheet View State")
        print(
            f"  Cursor pos (column, row): ({self.cursor_column}, {self.cursor_row}) or '{mCalc.Calc.get_cell_str(col=self.cursor_column, row=self.cursor_row)}'"
        )
        if self.column_split_mode == 1 and self.row_split_mode == 1:
            print(
                f"  Sheet is split vertically and horizontally at {self.vertical_split} / {self.horizontal_split}"
            )
        elif self.column_split_mode == 1:
            print(f"  Sheet is split vertically at {self.vertical_split}")
        elif self.row_split_mode == 1:
            print(f"  Sheet is split horizontally at {self.horizontal_split}")
        else:
            print("  Sheet is not split")

        print(f"  Number of focused pane: {self.pane_focus_num}")
        print(f"  Left column indicies of left/right panes: {self.column_left_pane} / {self.column_right_pane}")
        print(f"  Top row indicies of upper/lower panes: {self.row_upper_pane} / {self.row_lower_pane}")
        print()

    def to_string(self) -> str:
        """
        Gets string Representation of object.
        
        String representation can also be used to create a new instance of this class.
        
        same as ``str(instance)``
        """
        lst = [
            self.cursor_column,
            self.cursor_row,
            self.column_split_mode,
            self.row_split_mode,
            self.vertical_split,
            self.horizontal_split,
            self.pane_focus_num,
            self.column_left_pane,
            self.column_right_pane,
            self.row_upper_pane,
            self.row_lower_pane,
        ]
        return "/".join([str(val) for val in lst])

    def __str__(self) -> str:
        return self.to_string()
